# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Dict
from typing import Optional
from typing import Union

from kolena._utils.pydantic_v1 import StrictBool
from kolena._utils.pydantic_v1 import StrictFloat
from kolena._utils.pydantic_v1 import StrictInt
from kolena._utils.pydantic_v1 import StrictStr
from kolena._utils.pydantic_v1.dataclasses import dataclass


class EventAPI:
    class Path(str, Enum):
        PROFILE = "/event/profile"
        EVENT = "/event"

    class Event(str, Enum):
        # auth
        INITIALIZE_SDK_CLIENT = "sdk-client-initialized"

        # test case
        CREATE_TEST_CASE = "sdk-test-case-created"
        LOAD_TEST_CASE = "sdk-test-case-loaded"
        LOAD_TEST_CASE_SAMPLES = "sdk-test-case-samples-loaded"
        EDIT_TEST_CASE = "sdk-test-case-edited"
        INIT_MANY_TEST_CASES = "sdk-many-test-cases-initialized"

        # test suite
        CREATE_TEST_SUITE = "sdk-test-suite-created"
        LOAD_TEST_SUITE = "sdk-test-suite-loaded"
        LOAD_ALL_TEST_SUITES = "sdk-all-test-suites-loaded"
        EDIT_TEST_SUITE = "sdk-test-suite-edited"
        LOAD_TEST_SUITE_SAMPLES = "sdk-test-suite-samples-loaded"

        # model
        CREATE_MODEL = "sdk-model-created"
        LOAD_MODEL = "sdk-model-loaded"
        LOAD_ALL_MODEL = "sdk-all-models-loaded"

        # test run
        EXECUTE_TEST_RUN = "sdk-test-run-executed"

        # dataset
        REGISTER_DATASET = "sdk-dataset-registered"
        FETCH_DATASET = "sdk-dataset-fetched"
        FETCH_DATASET_HISTORY = "sdk-dataset-history-fetched"
        LIST_DATASETS = "sdk-datasets-listed"

        # dataset evaluation
        FETCH_DATASET_MODEL_RESULT = "sdk-dataset-model-result-fetched"
        UPLOAD_DATASET_MODEL_RESULT = "sdk-dataset-model-result-uploaded"

        # quality-standard
        FETCH_QUALITY_STANDARD_RESULT = "sdk-quality-standard-result-fetched"

    @dataclass(frozen=True)
    class RecordEventRequest:
        event_name: str
        additional_metadata: Optional[Dict[str, Union[StrictInt, StrictFloat, StrictStr, StrictBool, None]]] = None



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Optional
from typing import Type

import pandas as pd
import pandera as pa
from pandera.typing import Series

from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.datatypes import LoadableDataFrame
from kolena._utils.serde import as_deserialized_json
from kolena._utils.serde import as_serialized_json
from kolena._utils.serde import with_serialized_columns


_SCALAR_TYPES = [str, bool, int, float]
JSONObject = object


class TestSampleDataFrameSchema(pa.DataFrameModel):
    """General-purpose frame used for test samples in isolation or paired with ground truths and/or inferences."""

    test_sample: Series[JSONObject] = pa.Field(coerce=True)
    test_sample_metadata: Optional[Series[JSONObject]] = pa.Field(coerce=True)
    ground_truth: Optional[Series[JSONObject]] = pa.Field(coerce=True, nullable=True)
    inference: Optional[Series[JSONObject]] = pa.Field(coerce=True)


class TestSampleDataFrame(LoadableDataFrame[TestSampleDataFrameSchema]):
    @classmethod
    def get_schema(cls) -> Type[TestSampleDataFrameSchema]:
        return TestSampleDataFrameSchema

    def as_serializable(self) -> pd.DataFrame:
        return TestSampleDataFrame._serde(self, True)

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestSampleDataFrame":
        df_deserialized = TestSampleDataFrame._serde(df, False)
        df_validated = validate_df_schema(df_deserialized, TestSampleDataFrameSchema, trusted=True)
        return cast(TestSampleDataFrame, df_validated)

    @staticmethod
    def _serde(df: pd.DataFrame, serialize: bool) -> pd.DataFrame:
        serde_function = as_serialized_json if serialize else as_deserialized_json
        df_out = df.copy()
        df_out["test_sample"] = df_out["test_sample"].apply(serde_function)
        if "test_sample_metadata" in df.columns:
            df_out["test_sample_metadata"] = df_out["test_sample_metadata"].apply(serde_function)
        if "ground_truth" in df.columns:
            df_out["ground_truth"] = df_out["ground_truth"].apply(serde_function)
        if "inference" in df.columns:
            df_out["inference"] = df_out["inference"].apply(serde_function)
        return df_out


class TestSuiteTestSamplesDataFrameSchema(pa.DataFrameModel):
    """Data frame used for loading test samples grouped by test case."""

    test_case_id: Optional[Series[pa.typing.Int64]] = pa.Field(coerce=True)
    test_sample: Series[JSONObject] = pa.Field(coerce=True)
    test_sample_metadata: Series[JSONObject] = pa.Field(coerce=True)


class TestSuiteTestSamplesDataFrame(LoadableDataFrame[TestSuiteTestSamplesDataFrameSchema]):
    @classmethod
    def get_schema(cls) -> Type[TestSuiteTestSamplesDataFrameSchema]:
        return TestSuiteTestSamplesDataFrameSchema

    def as_serializable(self) -> pd.DataFrame:
        return TestSuiteTestSamplesDataFrame._serde(self, True)

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestSuiteTestSamplesDataFrame":
        df_deserialized = TestSuiteTestSamplesDataFrame._serde(df, False)
        df_validated = validate_df_schema(df_deserialized, TestSuiteTestSamplesDataFrameSchema, trusted=True)
        return cast(TestSuiteTestSamplesDataFrame, df_validated)

    @staticmethod
    def _serde(df: pd.DataFrame, serialize: bool) -> pd.DataFrame:
        serde_function = as_serialized_json if serialize else as_deserialized_json
        df_out = df.copy()
        df_out["test_sample"] = df_out["test_sample"].apply(serde_function)
        df_out["test_sample_metadata"] = df_out["test_sample_metadata"].apply(serde_function)
        return df_out


class TestCaseEditorDataFrameSchema(pa.DataFrameModel):
    test_case_name: Series[pa.typing.String] = pa.Field(nullable=True)
    test_sample_type: Series[pa.typing.String] = pa.Field(coerce=True)
    test_sample: Series[JSONObject] = pa.Field(coerce=True)  # TODO: validators?
    test_sample_metadata: Series[JSONObject] = pa.Field(coerce=True)
    ground_truth: Series[JSONObject] = pa.Field(coerce=True, nullable=True)
    remove: Series[pa.typing.Bool] = pa.Field(coerce=True)


class TestCaseEditorDataFrame(LoadableDataFrame[TestCaseEditorDataFrameSchema]):
    def as_serializable(self) -> pd.DataFrame:
        object_columns = ["test_sample", "test_sample_metadata", "ground_truth"]
        return with_serialized_columns(self, object_columns)

    @classmethod
    def get_schema(cls) -> Type[TestCaseEditorDataFrameSchema]:
        return TestCaseEditorDataFrameSchema

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestCaseEditorDataFrame":
        df_deserialized = df.copy()
        df_deserialized["test_sample"] = df["test_sample"].apply(as_deserialized_json)
        df_deserialized["test_sample_metadata"] = df["test_sample_metadata"].apply(as_deserialized_json)
        df_deserialized["ground_truths"] = df["ground_truths"].apply(as_deserialized_json)
        df_validated = validate_df_schema(df_deserialized, TestCaseEditorDataFrameSchema, trusted=True)
        return cast(TestCaseEditorDataFrame, df_validated)


class MetricsDataFrameSchema(pa.DataFrameModel):
    test_sample: Optional[Series[JSONObject]] = pa.Field(coerce=True)
    test_case_id: Optional[Series[pa.typing.Int64]] = pa.Field(coerce=True)
    configuration_display_name: Optional[Series[pa.typing.String]] = pa.Field(coerce=True, nullable=True)
    metrics: Series[JSONObject] = pa.Field(coerce=True)


class MetricsDataFrame(LoadableDataFrame[MetricsDataFrameSchema]):
    @classmethod
    def get_schema(cls) -> Type[MetricsDataFrameSchema]:
        return MetricsDataFrameSchema

    def as_serializable(self) -> pd.DataFrame:
        return MetricsDataFrame._serde(self, True)

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "MetricsDataFrame":
        df_deserialized = MetricsDataFrame._serde(df, False)
        df_validated = validate_df_schema(df_deserialized, MetricsDataFrameSchema, trusted=True)
        return cast(MetricsDataFrame, df_validated)

    @staticmethod
    def _serde(df: pd.DataFrame, serialize: bool) -> pd.DataFrame:
        serde_function = as_serialized_json if serialize else as_deserialized_json
        df_out = df.copy()
        if "test_sample" in df.columns:
            df_out["test_sample"] = df_out["test_sample"].apply(serde_function)
        df_out["metrics"] = df_out["metrics"].apply(serde_function)
        return df_out



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Assets are additional files that can be linked to the datapoints in your datasets.
Assets can be visualized in the Kolena when exploring your datasets, test cases, or model results.

The following asset types are available:

- [`ImageAsset`][kolena.asset.ImageAsset]
- [`PlainTextAsset`][kolena.asset.PlainTextAsset]
- [`BinaryAsset`][kolena.asset.BinaryAsset]
- [`PointCloudAsset`][kolena.asset.PointCloudAsset]
- [`VideoAsset`][kolena.asset.VideoAsset]
- [`AudioAsset`][kolena.asset.AudioAsset]

"""
from abc import ABCMeta
from typing import Optional

from kolena._utils.datatypes import DataCategory
from kolena._utils.datatypes import DataType
from kolena._utils.datatypes import TypedDataObject
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena._utils.validators import ValidatorConfig


class _AssetType(DataType):
    IMAGE = "IMAGE"
    PLAIN_TEXT = "PLAIN_TEXT"
    BINARY = "BINARY"
    POINT_CLOUD = "POINT_CLOUD"
    VIDEO = "VIDEO"
    AUDIO = "AUDIO"

    @staticmethod
    def _data_category() -> DataCategory:
        return DataCategory.ASSET


@dataclass(frozen=True, config=ValidatorConfig)
class Asset(TypedDataObject[_AssetType], metaclass=ABCMeta):
    """Base class for all asset types."""


@dataclass(frozen=True, config=ValidatorConfig)
class ImageAsset(Asset):
    """An image in a cloud bucket."""

    locator: str
    """The location of this image in a cloud bucket, e.g. `s3://my-bucket/path/to/my-image-asset.png`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.IMAGE


@dataclass(frozen=True, config=ValidatorConfig)
class PlainTextAsset(Asset):
    """A plain text file in a cloud bucket."""

    locator: str
    """The location of this text file in a cloud bucket, e.g. `s3://my-bucket/path/to/my-text-asset.txt`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.PLAIN_TEXT


@dataclass(frozen=True, config=ValidatorConfig)
class BinaryAsset(Asset):
    """A binary file in a cloud bucket."""

    locator: str
    """The location of this text file in a cloud bucket, e.g. `s3://my-bucket/path/to/my-binary-asset.bin`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.BINARY


@dataclass(frozen=True, config=ValidatorConfig)
class PointCloudAsset(Asset):
    """
    A three-dimensional point cloud located in a cloud bucket. Points are assumed to be specified in a right-handed,
    Z-up coordinate system with the origin around the sensor that captured the point cloud.
    """

    locator: str
    """The location of this point cloud in a cloud bucket, e.g. `s3://my-bucket/path/to/my-point-cloud.pcd`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.POINT_CLOUD


# NOTE: declare BaseVideoAsset as separate class for extension -- default fields in main VideoAsset class prevent
# extension with non-default fields
@dataclass(frozen=True, config=ValidatorConfig)
class BaseVideoAsset(Asset):
    """A video clip located in a cloud bucket or served at a URL."""

    locator: str
    """URL (e.g. S3, HTTPS) of the video file."""

    @classmethod
    def _data_type(cls) -> _AssetType:
        return _AssetType.VIDEO


@dataclass(frozen=True, config=ValidatorConfig)
class VideoAsset(BaseVideoAsset):
    """A video clip located in a cloud bucket or served at a URL."""

    locator: str
    """URL (e.g. S3, HTTPS) of the video file."""

    thumbnail: Optional[ImageAsset] = None
    """Optionally provide asset locator for custom video thumbnail image."""

    start: Optional[float] = None
    """Optionally specify start time of video snippet, in seconds."""

    end: Optional[float] = None
    """Optionally specify end time of video snippet, in seconds."""

    def __post_init__(self) -> None:
        if self.start is not None and self.end is not None and self.start > self.end:
            raise ValueError(f"Specified start time '{self.start}' is after specified end time '{self.end}'")
        if self.start is not None and self.end is not None and (self.start < 0 or self.end < 0):
            raise ValueError(f"Specified start time '{self.start}' and end time '{self.end}' must be non-negative")


@dataclass(frozen=True, config=ValidatorConfig)
class AudioAsset(Asset):
    """
    An audio file in a cloud bucket or served at a URL.

    For best results, use a broadly supported file type such as `.mp3` or `.wav`.
    """

    locator: str
    """The location of this audio file in a cloud bucket, e.g. `s3://my-bucket/path/to/my-audio-asset.mp3`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.AUDIO


_ASSET_TYPES = [ImageAsset, PlainTextAsset, BinaryAsset, PointCloudAsset, BaseVideoAsset, VideoAsset, AudioAsset]



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Tuple
from typing import Type

from kolena._utils.pydantic_v1 import validate_arguments
from kolena._utils.validators import ValidatorConfig
from kolena.workflow import GroundTruth
from kolena.workflow import Inference
from kolena.workflow import Model
from kolena.workflow import TestCase
from kolena.workflow import TestSample
from kolena.workflow import TestSuite
from kolena.workflow import Workflow


@validate_arguments(config=ValidatorConfig)
def define_workflow(
    name: str,
    test_sample_type: Type[TestSample],
    ground_truth_type: Type[GroundTruth],
    inference_type: Type[Inference],
) -> Tuple[Workflow, Type[TestCase], Type[TestSuite], Type[Model]]:
    """
    Define a new workflow, specifying its test sample, ground truth, and inference types.

    ```python
    from kolena.workflow import define_workflow

    from my_code import MyTestSample, MyGroundTruth, MyInference

    _, TestCase, TestSuite, Model = define_workflow(
        "My Workflow",
        MyTestSample,   # extends e.g. kolena.workflow.Image (or uses directly)
        MyGroundTruth,  # extends kolena.workflow.GroundTruth
        MyInference,    # extends kolena.workflow.Inference
    )
    ```

    `define_workflow` is provided as a convenience method to create the [`TestCase`][kolena.workflow.TestCase],
    [`TestSuite`][kolena.workflow.TestSuite], and [`Model`][kolena.workflow.Model] objects
    for a new workflow. These objects can also be defined manually by subclassing them and binding the `workflow`
    class variable:

    ```python
    from kolena.workflow import TestCase

    from my_code import my_workflow

    class MyTestCase(TestCase):
        workflow = my_workflow
    ```

    :param name: The name of the workflow.
    :param test_sample_type: The type of the [`TestSample`][kolena.workflow.TestSample] for this workflow.
    :param ground_truth_type: The type of the [`GroundTruth`][kolena.workflow.GroundTruth] for this workflow.
    :param inference_type: The type of the [`Inference`][kolena.workflow.Inference] for this workflow.
    :return: The `Workflow` object for this workflow along with the [`TestCase`][kolena.workflow.TestCase],
        [`TestSuite`][kolena.workflow.TestSuite], and [`Model`][kolena.workflow.Model] objects to use when creating and
        running tests for this workflow.
    """
    workflow = Workflow(
        name=name,
        test_sample_type=test_sample_type,
        ground_truth_type=ground_truth_type,
        inference_type=inference_type,
    )

    test_case = type("TestCase", (TestCase,), {"workflow": workflow})
    test_suite = type("TestSuite", (TestSuite,), {"workflow": workflow, "_test_case_type": test_case})
    model = type("Model", (Model,), {"workflow": workflow})

    return workflow, cast(Type[TestCase], test_case), cast(Type[TestSuite], test_suite), cast(Type[Model], model)



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The output from a [`Model`][kolena.workflow.Model]. In other words, a model is a deterministic transformation from a
[`TestSample`][kolena.workflow.TestSample] to an [`Inference`][kolena.workflow.Inference].

```python
from dataclasses import dataclass
from typing import Optional

from kolena.workflow import Inference
from kolena.workflow.annotation import Keypoints

@dataclass(frozen=True)
class PoseEstimate(Inference):
    skeleton: Optional[Keypoints] = None  # leave empty if nothing is detected
    confidence: Optional[float] = None
```
"""
from typing import Type

from kolena._utils.datatypes import DataObject
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena._utils.validators import ValidatorConfig
from kolena.workflow import Composite
from kolena.workflow import TestSample
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import safe_issubclass
from kolena.workflow._validators import validate_data_object_type
from kolena.workflow._validators import validate_field
from kolena.workflow.test_sample import _get_composite_fields


@dataclass(frozen=True, config=ValidatorConfig)
class Inference(DataObject):
    """
    The inference produced by a model.

    Typically the structure of this object closely mirrors the structure of the
    [`GroundTruth`][kolena.workflow.GroundTruth] for a workflow, but this is not a requirement.

    During evaluation, the [`TestSample`][kolena.workflow.TestSample] objects, ground truth objects, and these inference
    objects are provided to the [`Evaluator`][kolena.workflow.Evaluator] implementation to compute metrics.

    This object may contain any combination of scalars (e.g. `str`, `float`),
    [`Annotation`][kolena.workflow.annotation.Annotation] objects, or lists of these objects.

    A model processing a [`Composite`][kolena.workflow.Composite] test sample can produce an inference result for each
    of its elements. To associate an inference result to each test sample element, put the attributes and/or annotations
    inside a `DataObject` and use the same attribute name as that used in the [`Composite`][kolena.workflow.Composite]
    test sample.

    Continue with the example given in [`Composite`][kolena.workflow.Composite], where the `FacePairSample` test sample
    type is defined using a pair of images under the `source` and `target` members, we can design a corresponding
    inference type with image-level annotations defined in the `FaceRegion` object:

    ```python
    from dataclasses import dataclass

    from kolena.workflow import DataObject, Inference
    from kolena.workflow.annotation import BoundingBox, Keypoints

    @dataclass(frozen=True)
    class FaceRegion(DataObject):
        bounding_box: BoundingBox
        keypoints: Keypoints

    @dataclass(frozen=True)
    class FacePair(Inference):
        source: FaceRegion
        target: FaceRegion
        similarity: float
    ```

    This way, it is clear which bounding boxes and keypoints are associated to which image in the test sample.
    """


def _validate_inference_type(test_sample_type: Type[TestSample], inference_type: Type[Inference]) -> None:
    if not issubclass(inference_type, Inference):
        raise ValueError(f"Inference must subclass {Inference.__name__}")

    is_composite = issubclass(test_sample_type, Composite)
    composite_fields = _get_composite_fields(test_sample_type) if is_composite else []

    for field_name, field_value in get_data_object_field_types(inference_type).items():
        if field_name in composite_fields and safe_issubclass(field_value, DataObject):
            validate_data_object_type(field_value)
        else:
            validate_field(field_name, field_value)
